# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Server',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(max_length=256, verbose_name='server name', blank=True)),
                ('url', models.URLField(help_text=b'The URL, that the guardian API is available under.', verbose_name='API URL')),
                ('token', models.CharField(help_text=b'Add this to your client server settings as "SERVER_GUARDIAN_SECURITY_TOKEN".', max_length=256, verbose_name='token')),
                ('response_body', models.TextField(verbose_name='server response', blank=True)),
                ('status_code', models.PositiveIntegerField(null=True, verbose_name='server response status code', blank=True)),
                ('last_updated', models.DateTimeField(null=True, verbose_name='last updated', blank=True)),
            ],
        ),
    ]



"""Views for the ``server_guardian`` app."""
import os
import subprocess

from django.conf import settings
from django.contrib.auth.decorators import user_passes_test
from django.core.urlresolvers import reverse
from django.shortcuts import redirect
from django.utils.decorators import method_decorator
from django.views.generic import ListView, View

from django_libs.views_mixins import JSONResponseMixin

from . import models
from .default_settings import DASHBOARD_VIEW_PERMISSION
from .management.commands.guardian_fetch import LOCKFILE


class GuardianDashboardView(ListView):
    """View that lists all servers and their status"""
    model = models.Server

    @method_decorator(user_passes_test(DASHBOARD_VIEW_PERMISSION))
    def dispatch(self, request, *args, **kwargs):
        return super(GuardianDashboardView, self).dispatch(
            request, *args, **kwargs)


class GuardianReloadView(JSONResponseMixin, View):
    """View, that allows reloading server status via ajax."""

    @method_decorator(user_passes_test(DASHBOARD_VIEW_PERMISSION))
    def dispatch(self, request, *args, **kwargs):
        if not request.is_ajax():
            return redirect(reverse('server_guardian_dashboard'))
        return super(GuardianReloadView, self).dispatch(request, *args,
                                                        **kwargs)

    def get(self, request, *args, **kwargs):
        context = self.get_context_data(**kwargs)
        return self.render_to_response(context)

    def post(self, request, *args, **kwargs):
        context = self.get_context_data(**kwargs)
        return self.render_to_response(context)

    def get_context_data(self, **kwargs):
        ctx = {}
        if os.path.exists(LOCKFILE):
            # if lockfile exists, we know the command is still running
            ctx.update({
                'STATUS': 'running',
                'MESSAGE': 'Update command is still running. Please wait...',
            })
        elif self.request.method == 'GET':
            # GET requests are only to update the UI
            ctx.update({
                'STATUS': 'finished',
                'MESSAGE': 'Update command finished.',
            })
        elif self.request.method == 'POST':
            # POST requests are meant to start the command
            # if it is already running, the lockfile check above is fired
            status, message = self.start_command()
            ctx.update({'STATUS': status, 'MESSAGE': message})
        return ctx

    def start_command(self):

        # TODO not the most beautiful code, but it allows for easy reporting
        try:
            command = 'guardian_fetch'
            project_root = getattr(
                settings, 'DJANGO_PROJECT_ROOT', settings.PROJECT_ROOT)
            manage_py = os.path.join(project_root, 'manage.py')
            venv = os.environ.get('VIRTUAL_ENV', None)
            python = '/usr/bin/python'
            if venv is not None:
                python = os.path.join(venv, 'bin/python')
            popen_args = [python, manage_py, command]
            subprocess.Popen(popen_args)
        except Exception as ex:
            return 'error', (
                'There was an error while starting the update command:'
                ' {0}'.format(ex)
            )
        return 'started', 'Update command has been started. Please wait...'



"""Tests for the models of the server_guardian app."""
from django.test import TestCase
from mixer.backend.django import mixer
from server_guardian import constants


class ServerTestCase(TestCase):
    """Tests for the ``Server`` model class."""
    longMessage = True

    def test_instantiation(self):
        server = mixer.blend('server_guardian.Server')
        self.assertTrue(server.pk)

    def test_get_response_dict(self):
        json_response = (
            '[{"label": "foobar", "status": "OK", "info": "it is OK"}]'
        )
        server = mixer.blend('server_guardian.Server',
                             response_body=json_response)
        self.assertEqual(
            server.get_parsed_response()[0]['status'],
            constants.SERVER_STATUS['OK'],
            msg=(
                'It should be possible to read the status from'
                ' the response dict.'
            )
        )
        self.assertEqual(
            server.get_parsed_response()[0]['info'],
            'it is OK',
            msg=(
                'It should be possible to read the info from the'
                ' response dict.'
            )
        )
        html_response = '<html><head></head><body>HTML RESPONSE</body></html>'
        server2 = mixer.blend('server_guardian.Server',
                              response_body=html_response)
        self.assertEqual(
            server2.get_parsed_response()[0]['status'],
            constants.SERVER_STATUS['DANGER'],
            msg=(
                'For an HTML response, the dict should return a'
                ' warning status.'
            )
        )

    def test_has_errors(self):
        # no error status
        json_response = (
            '[{"label": "foobar", "status": "OK", "info": "it is OK"}]'
        )
        server = mixer.blend('server_guardian.Server',
                             response_body=json_response)
        self.assertFalse(
            server.has_errors(),
            msg='The server should have no errors.')

        # error status
        json_response = (
            '[{"label": "foobar", "status": "OK", "info": "it is OK"}'
            '{"label": "barfoo", "status": "WARNING", "info": "oops"]'
        )
        server.response_body = json_response
        self.assertTrue(
            server.has_errors(),
            msg='The server should have an error.')

        # empty body and status code of 404
        json_response = ''
        server.response_body = json_response
        server.status_code = 404
        self.assertTrue(
            server.has_errors(),
            msg='The server should have an error.')



"""The command, that fetches the information from the configured servers."""
import os
import sys

from django.conf import settings
from django.core.management.base import BaseCommand
from django.utils.timezone import now

import requests
from django_libs.utils_email import send_email
from django_libs.decorators import lockfile

from ... import default_settings
from ...models import Server
from ...constants import SERVER_STATUS

LOCKFILE = os.path.join(settings.LOCKFILE_FOLDER, 'guardian_fetch')


class Command(BaseCommand):

    def send_error_email(self, server):
        context = {
            'server': server,
            'subject': u'{} - {}'.format(server.status_code, server.name),
        }
        send_email(
            request={},
            extra_context=context,
            subject_template='server_guardian/email/email_subject.html',
            body_template='server_guardian/email/warning_email_body.html',
            from_email=settings.FROM_EMAIL,
            recipients=[admin[1] for admin in settings.ADMINS],
        )

    def get_email_required(self, server):
        """
        Send a warning email, if the server response contains errors.

        This includes when the server reports a status of "DANGER" or if the
        json response failed to decode or if the response code was within the
        EMAIL_ON_STATUS settings.

        """
        if server.status_code in default_settings.EMAIL_ON_STATUS:
            return True

        try:
            for metric in server.get_parsed_response():
                if metric['status'] == SERVER_STATUS['DANGER']:
                    return True
        except (KeyError, TypeError) as ex:
            server.response_body = (
                'Server got an error "{0}" and returned the following content:'
                '{1}'.format(ex, server.response_body)
            )
            server.save()
            return True

        return False

    @lockfile(LOCKFILE)
    def handle(self, *args, **options):

        count = 0
        mails_sent = 0

        for server in Server.objects.all():
            response = requests.get(url=server.url, params={
                'token': server.token
            })
            server.response_body = response.content
            server.status_code = response.status_code
            server.last_updated = now()
            server.save()
            if self.get_email_required(server):
                self.send_error_email(server)
                mails_sent += 1
            count += 1

        sys.stdout.write(
            '[{0}] Updated {1} servers. Sent {2} warning emails.\n'.format(
                now(), count, mails_sent)
        )



"""Collection of constants fo the ``server_guardian`` app."""


# keep this in sync with server_guardian_api.constants
SERVER_STATUS = {  # pragma: no cover
    'OK': 'OK',
    'WARNING': 'WARNING',
    'DANGER': 'DANGER',
}

ERROR_STATUS = [SERVER_STATUS['WARNING'], SERVER_STATUS['DANGER']]  # pragma: no cover  # NOQA

HTML_STATUS_FALLBACK = [{  # pragma: no cover
    'status': SERVER_STATUS['DANGER'],
    'info': 'JSON response could not be decoded.'
}]



"""Tests for the ``guardian_fetch`` command."""
import requests
from collections import namedtuple

from django.core import mail
from django.core.management import call_command
from django.test import TestCase

import mock

from .mixes import new_server_factory
from ..constants import SERVER_STATUS
from ..models import Server


class CommandTestCase(TestCase):
    """Test case for the ``guardian_fetch`` command."""
    longMessage = True

    def setUp(self):
        self.server = new_server_factory()
        self.server_response = namedtuple(
            'response',
            ['status_code', 'content']
        )
        self.get_mock = requests.get

    def tearDown(self):
        requests.get = self.get_mock

    def test_command(self):
        self.server_response.status_code = 200
        json_response = (
            '[{"label": "foobar", "status": "OK", "info": "it is OK"}]'
        )
        self.server_response.content = json_response
        requests.get = mock.MagicMock(
            return_value=self.server_response,
        )
        call_command('guardian_fetch')
        server = Server.objects.get(pk=self.server.pk)
        self.assertEqual(
            server.get_parsed_response()[0]['status'],
            SERVER_STATUS['OK'],
            msg=(
                'After the command was run, the server instance should have'
                ' an updated status of "OK".'
            )
        )

        self.server_response.status_code = 404
        requests.get = mock.MagicMock(
            return_value=self.server_response,
        )
        call_command('guardian_fetch')
        self.assertEqual(len(mail.outbox), 1, msg=(
            'There should be one email sent, because the server responsed with'
            ' a code of 404.'
        ))

        mail.outbox = []
        self.server_response.status_code = 200
        self.server_response.content = 'this is not JSON'
        requests.get = mock.MagicMock(
            return_value=self.server_response,
        )
        call_command('guardian_fetch')
        self.assertEqual(len(mail.outbox), 1, msg=(
            'There should be one email sent if the response is not JSON.'
        ))

        mail.outbox = []
        self.server_response.status_code = 200
        self.server_response.content = '{"wrong": "JSON data"}'
        requests.get = mock.MagicMock(
            return_value=self.server_response,
        )
        call_command('guardian_fetch')
        self.assertEqual(len(mail.outbox), 1, msg=(
            'There should be one email sent if the response is JSON, but in'
            ' the wrong format.'
        ))



"""Pre-made model factories with mixer."""
from mixer.backend.django import mixer


def new_server_factory():
    """Returns a Server instance like it was just set up."""
    return mixer.blend(
        'server_guardian.Server',
        status_code=None,
        response_body='',
        last_updated=None,
    )



"""Admin classes for the server_guardian app."""
from django.contrib import admin

from . import models


class ServerAdmin(admin.ModelAdmin):
    """Custom admin for the ``Server`` model."""
    list_display = ('name', 'url', 'status_code')
    fields = ('name', 'url', 'token')
    search_fields = ['name']


admin.site.register(models.Server, ServerAdmin)



"""Default settings for the app."""
from django.conf import settings

EMAIL_ON_STATUS = getattr(
    settings,
    'SERVER_GUARDIAN_EMAIL_ON_STATUS',
    [403, 404, 405],
)


DASHBOARD_VIEW_PERMISSION = getattr(
    settings,
    'SERVER_GUARDIAN_DASHBOARD_VIEW_PERMISSION',
    lambda u: u.is_superuser
)



"""URLs for the server_guardian app."""
from compat import patterns, url

from . import views


urlpatterns = patterns(
    '',
    url(r'^$',
        views.GuardianDashboardView.as_view(),
        name='server_guardian_dashboard'),
    url(r'^reload/$',
        views.GuardianReloadView.as_view(),
        name='server_guardian_reload'),
)
